# Copyright 2013 Openstack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import json
from simplejson import JSONDecodeError

from webob.dec import wsgify

from wafflehaus.base import WafflehausBase
import wafflehaus.resource_filter as rf


class EditResponse(WafflehausBase):
    def __init__(self, app, conf):
        super(EditResponse, self).__init__(app, conf)
        self.log.name = conf.get('log_name', __name__)
        self.log.info('Starting wafflehaus edit_response middleware')
        self.resources = {}
        filters = conf.get('filters')
        if filters is None:
            self.log.warning("EditResponse waffle could not find any filters"
                             " in its configuration")
            return
        for resource_filter in filters.split():
            if resource_filter in self.resources.keys():
                self.log.warning("EditResponse waffle found two filter names "
                                 "with the same name (first now overridden")
            self.resources[resource_filter] = {
                "resource": rf.parse_resources(
                    conf.get("%s_resource" % resource_filter)),
                "key": conf.get("%s_key" % resource_filter),
                "value": conf.get("%s_value" % resource_filter)}
        return

    def _replace_lookup(self, replace_str):
        if replace_str == '[]':
            return []
        if replace_str.lower() == 'null' or replace_str.lower() == 'none':
            return None
        if replace_str == '{}':
            return {}
        return replace_str

    def _foreach(self, conditional, data):
        if not isinstance(data, list):
            return data
        if not all(isinstance(item, dict) for item in data):
            return data
        new_data = []
        splits = conditional.split(":", 2)
        action = splits[1]
        conditionals = splits[2].split(',')
        for item in data:
            do_action = False
            for cond in conditionals:
                # TODO(jlh): Probably should support more than '='
                (target, value) = cond.split("=")
                if target not in item:
                    continue
                if item.get(target) == value:
                    do_action = True
            if do_action:
                if action == 'keep_if':
                    new_data.append(item)
            if not do_action:
                if action == 'drop_if':
                    new_data.append(item)
        self.log.debug('Replacing "{0}" with :"{1}"'.format(data, new_data))
        return new_data

    def _change_attribs(self, req, resp, resource):
        # Not sure recursion is the way here...
        def walk_keys(data):
            if isinstance(data, dict):
                for key, value in data.items():
                    if key == resource['key']:
                        val = resource.get('value', None)
                        if val is None:
                            del(data[key])
                            self.log.debug('Deleting "{0}":"{1}" from the '
                                           'response'.format(key, value))
                        elif val.startswith('foreach:'):
                            data[key] = self._foreach(val, data[key])
                        else:
                            replace = self._replace_lookup(resource['value'])
                            self.log.debug('Replacing "{0}":"{1}" with "{0}"'
                                           ':"{2}"'.format(key, value,
                                                           resource['value']))
                            data[key] = replace
                    else:
                        if isinstance(value, dict) or isinstance(value, list):
                            data[key] = walk_keys(value)
            elif isinstance(data, list):
                data = [walk_keys(part) for part in data]
            return data

        try:
            new_body = resp.json
            new_body = walk_keys(new_body)
            if 'http_status_code' in resource['key']:
                resource_val = resource['value']
                if resource_val.startswith('replace_if'):
                    (status_chk, status_repl) = resource_val.split(':')[1:3]
                    if str(resp.status_code) == status_chk:
                        self.log.debug('Replacing http status code "{0}" with '
                                       '"{1}"'.format(status_chk, status_repl))
                        resp.status_code = int(status_repl)
            resp.body = json.dumps(new_body)
        except JSONDecodeError:
            pass

        return resp

    @wsgify
    def __call__(self, req):
        """Returns a response if processed or an app if skipped."""
        super(EditResponse, self).__call__(req)
        resp = None

        if not self.enabled:
            return self.app
        if hasattr(self, "resources"):
            for resource_filter in self.resources.keys():
                if rf.matched_request(
                        req, self.resources[resource_filter]["resource"]):
                    if resp is None:
                        resp = req.get_response(self.app)
                    resp = self._change_attribs(
                        req, resp, self.resources[resource_filter])
        if resp is None:
            return self.app
        else:
            return resp


def filter_factory(global_conf, **local_conf):
    """Returns a WSGI filter app for use with paste.deploy."""
    conf = global_conf.copy()
    conf.update(local_conf)

    def wrapper(app):
        return EditResponse(app, conf)

    return wrapper



# Copyright 2013 Openstack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import webob.dec
import webob.exc

import wafflehaus.base


class ContextFilter(wafflehaus.base.WafflehausBase):
    """Attempts to create a context with the configured context class."""
    def __init__(self, app, conf):
        super(ContextFilter, self).__init__(app, conf)
        self.log.name = conf.get('log_name', __name__)
        self.log.info('Starting wafflehaus context middleware')
        self.context_key = conf.get('context_key')
        self.strat = conf.get('context_strategy')
        if self.strat is None:
            self.log.info('No context context is configured')
        self.req_auth = conf.get('require_auth_info') in self.truths

    def _import_class(self, name):
        last_dot = name.rfind(".")
        kls = name[last_dot + 1: len(name)]
        module = __import__(name[0:last_dot], globals(), locals(), [kls])
        return getattr(module, kls)

    def _create_context(self, req):
        if self.strat is None:
            return self.app
        try:
            kls = self._import_class(self.strat)
            if self.testing:
                self.log.info('Would be setting %s with %s' %
                              (self.context_key, self.strat))
        except ImportError:
            self.log.error("Could not find context strategy: %s" % self.strat)
            if not self.testing:
                return webob.exc.HTTPInternalServerError()
            else:
                self.log.error("Failed to find strategy: %s" % self.strat)
        if not self.testing:
            self.strat_instance = kls(self.context_key, req_auth=self.req_auth)
            result = self.strat_instance.load_context(req)
            if not result:
                return webob.exc.HTTPForbidden()
        return self.app

    @webob.dec.wsgify
    def __call__(self, req):
        super(ContextFilter, self).__call__(req)
        if not self.enabled:
            return self.app

        return self._create_context(req)


class BaseContextStrategy(object):

    def __init__(self, key, req_auth=False):
        self.key = key
        self.require_auth_info = req_auth

    def load_context(self, req):
        return True


def filter_factory(global_conf, **local_conf):
    """Returns a WSGI filter app for use with paste.deploy."""
    conf = global_conf.copy()
    conf.update(local_conf)

    def try_context(app):
        return ContextFilter(app, conf)
    return try_context



# Copyright 2013 Openstack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from wafflehaus.base import WafflehausBase
from wafflehaus import tests


class TestWafflehausBase(tests.TestCase):

    def test_create_default_instance(self):
        base = WafflehausBase(self.app, {})
        self.assertIsNotNone(base)
        self.assertFalse(base.testing)
        self.assertEqual(base.app, self.app)

    def test_create_default_instance_with_testing(self):
        for testing in (True, 'True', 'true', 't', '1', 'on', 'yes', 'y'):
            base = WafflehausBase(self.app, {'testing': testing})
            self.assertIsNotNone(base)
            self.assertTrue(base.testing)
            self.assertEqual(base.app, self.app)

    def test_create_default_instance_is_not_enabled(self):
        base = WafflehausBase(self.app, {})
        self.assertIsNotNone(base)
        self.assertEqual(base.app, self.app)
        self.assertFalse(base.enabled)

    def test_create_default_instance_configured_enabled(self):
        for enabled in (True, 'True', 'true', 't', '1', 'on', 'yes', 'y'):
            base = WafflehausBase(self.app, {'enabled': enabled})
            self.assertIsNotNone(base)
            self.assertEqual(base.app, self.app)
            self.assertTrue(base.enabled)

    def test_handle_header_override_called_when_configured(self):
        self.set_reconfigure()
        oc_path = 'wafflehaus.base.WafflehausBase._override'
        m_override_caller = self.create_patch(oc_path)
        base = WafflehausBase(self.app, {})
        base.__call__(None)
        self.assertTrue(m_override_caller.called)

    def test_handle_header_override_not_called_when_not_configured(self):
        oc_path = 'wafflehaus.base.WafflehausBase._override'
        m_override_caller = self.create_patch(oc_path)
        base = WafflehausBase(self.app, {})
        base.__call__(None)
        self.assertFalse(m_override_caller.called)



# Copyright 2013 Openstack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import mock

from wafflehaus.routers import rolerouter
from wafflehaus import tests


class TestRoleRouter(tests.TestCase):
    def setUp(self):
        super(TestRoleRouter, self).setUp()
        self.local_conf = {"enabled": "true", "routes": "cat dog",
                           "roles_cat": "domestic outdoor",
                           "roles_dog": "mutt",
                           "route_cat": "cat_filter cat_app",
                           "route_dog": "dog_filter dog_app",
                           "route_default": "appx"}
        self.key_conf = {"enabled": "true", "context_key": "animal.context",
                         "routes": "cat dog",
                         "roles_cat": "domestic outdoor",
                         "roles_dog": "mutt",
                         "route_cat": "cat_filter cat_app",
                         "route_dog": "dog_filter dog_app",
                         "route_default": "appx"}

        def named_mock(name):
            x = mock.Mock()
            x.name = name
            x.return_value = name
            return x

        self.loader = mock.Mock()
        self.loader.get_filter = named_mock
        self.loader.get_app = named_mock

        self.global_conf = {}

    def test_get_default_instance(self):
        result = rolerouter.rolerouter_factory(self.loader, self.global_conf,
                                               **self.local_conf)
        self.assertTrue(isinstance(result, rolerouter.RoleRouter))
        self.assertEqual(len(result.routes), 3)
        self.assertTrue(all(k in result.routes.keys()
                            for k in ["default", "cat", "dog"]))
        self.assertEqual(len(result.roles), 3)
        self.assertTrue(all(k in result.roles.keys()
                            for k in ["domestic", "outdoor", "mutt"]))

    def test_create_instance_with_key(self):
        result = rolerouter.rolerouter_factory(self.loader, self.global_conf,
                                               **self.key_conf)
        self.assertTrue(isinstance(result, rolerouter.RoleRouter))
        self.assertEqual(len(result.routes), 3)
        self.assertTrue(all(k in result.routes.keys()
                            for k in ["default", "cat", "dog"]))
        self.assertEqual(len(result.roles), 3)
        self.assertTrue(all(k in result.roles.keys()
                            for k in ["domestic", "outdoor", "mutt"]))
        self.assertEqual('animal.context', result.context_key)

    def test_call_to_domestic_role(self):
        context = mock.Mock()
        context.roles = ["domestic"]

        req = mock.Mock()
        req.environ = {"nova.context": context}

        rr = rolerouter.rolerouter_factory(self.loader, self.global_conf,
                                           **self.local_conf)
        result = rr(req)
        self.assertEqual(result, "cat_filter")

    def test_call_to_outdoor_role(self):
        context = mock.Mock()
        context.roles = ["outdoor"]

        req = mock.Mock()
        req.environ = {"nova.context": context}

        rr = rolerouter.rolerouter_factory(self.loader, self.global_conf,
                                           **self.local_conf)
        result = rr(req)
        self.assertEqual(result, "cat_filter")

    def test_call_to_mutt_role(self):
        context = mock.Mock()
        context.roles = ["mutt"]

        req = mock.Mock()
        req.environ = {"nova.context": context}

        rr = rolerouter.rolerouter_factory(self.loader, self.global_conf,
                                           **self.local_conf)
        result = rr(req)
        self.assertEqual(result, "dog_filter")

    def test_call_to_untracked_role(self):
        context = mock.Mock()
        context.roles = ["blah"]

        req = mock.Mock()
        req.environ = {"nova.context": context}

        rr = rolerouter.rolerouter_factory(self.loader, self.global_conf,
                                           **self.local_conf)
        result = rr(req)

        # NOTE(jmeridth/roaet): this needs to be called explicitly because
        # it doesn't get called in the filter chain
        self.assertEqual(result(), "appx")

    def test_multiple_roles_from_same_route(self):
        context = mock.Mock()
        context.roles = ["domestic", "outdoor"]

        req = mock.Mock()
        req.environ = {"nova.context": context}

        rr = rolerouter.rolerouter_factory(self.loader, self.global_conf,
                                           **self.local_conf)
        result = rr(req)
        self.assertEqual(result, "cat_filter")

    def test_ordering_multiple_roles_different_route_returns_cat_filter(self):
        context = mock.Mock()
        context.roles = ["domestic", "mutt"]

        req = mock.Mock()
        req.environ = {"nova.context": context}

        rr = rolerouter.rolerouter_factory(self.loader, self.global_conf,
                                           **self.local_conf)
        result = rr(req)
        self.assertEqual(result, "cat_filter")

    def test_ordering_of_roles_does_not_matter(self):
        context = mock.Mock()
        context.roles = ["mutt", "outdoor"]

        req = mock.Mock()
        req.environ = {"nova.context": context}

        rr = rolerouter.rolerouter_factory(self.loader, self.global_conf,
                                           **self.local_conf)
        result = rr(req)
        self.assertEqual(result, "cat_filter")



# Copyright 2013 Openstack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
from mock import patch
from oslo_config import cfg
import unittest2


CONF = cfg.CONF


class TestCase(unittest2.TestCase):
    '''Class to decide which unit test class to inherit from uniformly.'''
    def setUp(self):
        super(TestCase, self).setUp()
        CONF.WAFFLEHAUS.runtime_reconfigurable = False
        self.app = mock.Mock()

    def set_reconfigure(self):
        CONF.WAFFLEHAUS.runtime_reconfigurable = True

    def create_patch(self, name, func=None):
        patcher = patch(name)
        thing = patcher.start()
        self.addCleanup(patcher.stop)
        return thing



# Copyright 2013 Openstack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import json

import webob

from wafflehaus import edit_response
from wafflehaus import tests


class TestEditResponse(tests.TestCase):
    def setUp(self):
        super(TestEditResponse, self).setUp()
        self.app = self._fake_app
        self.body = {"result":
                     {"passcode": "123password",
                      "combination": "1,2,3,4",
                      "secret": "no pants",
                      "recipe": "raw chicken, razzle dazzle"}}
        self.combo_conf = {"enabled": "true",
                           "filters": "safe secret",
                           "safe_resource": "POST /data",
                           "safe_key": "combination",
                           "safe_value": "REDACTED",
                           "secret_resource": "POST /data",
                           "secret_key": "secret"}
        self.delete_conf = {"enabled": "true",
                            "filters": "super_secret",
                            "super_secret_resource": "PUT /secrets",
                            "super_secret_key": "passcode"}
        self.redact_conf = {"enabled": "true",
                            "filters": "secret",
                            "secret_resource": "GET /sauce",
                            "secret_key": "recipe",
                            "secret_value": "REDACTED"}
        self.make_empty = {"enabled": "true",
                           "filters": "result",
                           "result_resource": "GET /sauce",
                           "result_key": "result",
                           "result_value": "[]"}
        self.make_null = {"enabled": "true",
                          "filters": "result",
                          "result_resource": "GET /sauce",
                          "result_key": "result",
                          "result_value": "NULL"}
        self.keep_if = {"enabled": "true",
                        "filters": "result",
                        "result_resource": "GET /sauce",
                        "result_key": "results",
                        "result_value": "foreach:keep_if:some=here"}
        self.drop_if = {"enabled": "true",
                        "filters": "result",
                        "result_resource": "GET /sauce",
                        "result_key": "results",
                        "result_value": "foreach:drop_if:some=here"}
        self.test_none = {"enabled": "true",
                          "filters": "result",
                          "result_resource": "GET /sauce/{id}",
                          "result_key": "result",
                          "result_value": "NULL"}

    @webob.dec.wsgify
    def _fake_app(self, req, body=None):
        if body is None:
            body = self.body
        return webob.Response(body=json.dumps(body), status=200)

    def test_filter_creation(self):
        test_filter = edit_response.filter_factory(self.combo_conf)(self.app)

        self.assertIsNotNone(test_filter)
        self.assertIsInstance(test_filter, edit_response.EditResponse)
        self.assertTrue(callable(test_filter))

    def test_keep_if(self):
        app_body = {"results":
                    [{"some": "here"}, {"some": "there"}, {"derp": "derp"}]}
        processed_body = {"results":
                          [{"some": "there"}, {"derp": "derp"}]}

        @webob.dec.wsgify
        def app(req):
            return webob.Response(body=json.dumps(app_body), status=200)
        test_filter = edit_response.filter_factory(self.drop_if)(app)
        resp = test_filter(webob.Request.blank("/sauce", method="GET"))

        self.assertEqual(resp.json, processed_body)

    def test_drop_if(self):
        app_body = {"results":
                    [{"some": "here"}, {"some": "there"}, {"derp": "derp"}]}
        processed_body = {"results": [{"some": "here"}]}

        @webob.dec.wsgify
        def app(req):
            return webob.Response(body=json.dumps(app_body), status=200)
        test_filter = edit_response.filter_factory(self.keep_if)(app)
        resp = test_filter(webob.Request.blank("/sauce", method="GET"))

        self.assertEqual(resp.json, processed_body)

    def test_disabled_filter(self):
        conf = {"enabled": "false"}
        test_filter = edit_response.filter_factory(conf)(self.app)
        resp = test_filter(webob.Request.blank("/cheeseburger", method="GET"))

        self.assertEqual(resp, self.app)

    def test_attrib_rename(self):
        test_filter = edit_response.filter_factory(self.redact_conf)(self.app)
        resp = test_filter(webob.Request.blank("/sauce", method="GET"))
        new_body = self.body
        new_body["result"]["recipe"] = "REDACTED"

        self.assertEqual(resp.json, new_body)
        self.assertEqual(resp.status_code, 200)

    def test_attrib_deletion(self):
        test_filter = edit_response.filter_factory(self.delete_conf)(self.app)
        resp = test_filter(webob.Request.blank("/secrets", method="PUT"))
        new_body = self.body
        del(new_body['result']['passcode'])

        self.assertEqual(resp.json, new_body)
        self.assertEqual(resp.status_code, 200)

    def test_make_empty(self):
        test_filter = edit_response.filter_factory(self.make_empty)(self.app)
        resp = test_filter(webob.Request.blank("/sauce", method="GET"))
        new_body = json.dumps(dict(result=[]))
        resp_json = json.dumps(json.loads(resp.body))

        self.assertEqual(resp_json, new_body)
        self.assertEqual(resp.status_code, 200)

    def test_make_null(self):
        test_filter = edit_response.filter_factory(self.make_null)(self.app)
        resp = test_filter(webob.Request.blank("/sauce", method="GET"))
        new_body = json.dumps(dict(result=None))
        resp_json = json.dumps(json.loads(resp.body))

        self.assertEqual(resp_json, new_body)
        self.assertEqual(resp.status_code, 200)

    def test_attrib_combo(self):
        test_filter = edit_response.filter_factory(self.combo_conf)(self.app)
        resp = test_filter(webob.Request.blank("/data", method="POST"))
        new_body = self.body
        new_body['result']['combination'] = 'REDACTED'
        del(new_body['result']['secret'])

        self.assertEqual(resp.json, new_body)
        self.assertEqual(resp.status_code, 200)

    def test_resource_with_alternate_methods(self):
        test_filter = edit_response.filter_factory(self.combo_conf)(self.app)
        put_resp = test_filter(webob.Request.blank("/data", method="PUT"))
        get_resp = test_filter(webob.Request.blank("/data", method="GET"))
        head_resp = test_filter(webob.Request.blank("/data", method="HEAD"))

        self.assertEqual(put_resp, self.app)
        self.assertEqual(get_resp, self.app)
        self.assertEqual(head_resp, self.app)

    def test_url_garbage(self):
        test_filter = edit_response.filter_factory(self.combo_conf)(self.app)
        resp1 = test_filter(webob.Request.blank("/gibberish/data",
                                                method="POST"))
        resp2 = test_filter(webob.Request.blank("/hacks?stuff=/data",
                                                method="POST"))
        resp3 = test_filter(webob.Request.blank("/stuffanddata",
                                                method="POST"))

        self.assertEqual(resp1, self.app)
        self.assertEqual(resp2, self.app)
        self.assertEqual(resp3, self.app)

    def test_attrib_case_sensitivity(self):
        app_body = {"results":
                    {"Combination": "DONT LOOK",
                     "SECRET": "<encrypted text>"}}

        @webob.dec.wsgify
        def app(req):
            return webob.Response(body=json.dumps(app_body), status=200)

        test_filter = edit_response.filter_factory(self.combo_conf)(app)
        resp = test_filter(webob.Request.blank("/data", method="POST"))

        self.assertEqual(resp.json, app_body)

    def test_nested_lists_and_dicts(self):
        app_body = {"results":
                    [{"some": "here"},
                     {"some": "there"},
                     ["random_string",
                      {"secret": "MY SECRETS",
                       "nested":
                       {"combination": "1,2,3,4"}}]]}
        processed_body = {"results":
                          [{"some": "here"},
                           {"some": "there"},
                           ["random_string",
                            {"nested":
                             {"combination": "REDACTED"}}]]}

        @webob.dec.wsgify
        def app(req):
            return webob.Response(body=json.dumps(app_body), status=200)

        test_filter = edit_response.filter_factory(self.combo_conf)(app)
        resp = test_filter(webob.Request.blank("/data", method="POST"))

        self.assertEqual(resp.json, processed_body)

    def test_none_request(self):
        app_body = "garbage"

        @webob.dec.wsgify
        def app(req):
            return webob.Response(body=app_body, status=200)
        test_filter = edit_response.filter_factory(self.test_none)(app)
        resp = test_filter(webob.Request.blank("/sauce/None", method="GET"))

        self.assertEqual(resp.body, app_body)

    def test_http_status_replace_get(self):
        app_body = {"garbage": {"garbage": "here"}}

        test_status = {"enabled": "true",
                       "filters": "httpget",
                       "httpget_resource": "GET /sauce",
                       "httpget_key": "http_status_code",
                       "httpget_value": "replace_if:200:201"}

        @webob.dec.wsgify
        def app(req):
            return webob.Response(body=json.dumps(app_body), status=200)
        test_filter = edit_response.filter_factory(test_status)(app)

        resp = test_filter(webob.Request.blank("/sauce", method="GET"))
        self.assertEqual(resp.status_code, 201, resp)

    def test_http_status_replace_post(self):
        app_body = {"garbage": {"garbage": "here"}}

        test_status = {"enabled": "true",
                       "filters": "httppost",
                       "httppost_resource": "POST /sauce",
                       "httppost_key": "http_status_code",
                       "httppost_value": "replace_if:200:201"}

        @webob.dec.wsgify
        def app(req):
            return webob.Response(body=json.dumps(app_body), status=200)
        test_filter = edit_response.filter_factory(test_status)(app)

        resp = test_filter(webob.Request.blank("/sauce", method="POST"))
        self.assertEqual(resp.status_code, 201, resp)

    def test_http_status_replace_put(self):
        app_body = {"garbage": {"garbage": "here"}}

        test_status = {"enabled": "true",
                       "filters": "httpput",
                       "httpput_resource": "PUT /sauce/{id}",
                       "httpput_key": "http_status_code",
                       "httpput_value": "replace_if:200:201"}

        @webob.dec.wsgify
        def app(req):
            return webob.Response(body=json.dumps(app_body), status=200)
        test_filter = edit_response.filter_factory(test_status)(app)

        resp = test_filter(webob.Request.blank("/sauce/id", method="PUT"))
        self.assertEqual(resp.status_code, 201, resp)

    def test_http_status_replace_delete(self):
        app_body = {"garbage": {"garbage": "here"}}

        test_status = {"enabled": "true",
                       "filters": "httpdelete",
                       "httpdelete_resource": "DELETE /sauce/{id}",
                       "httpdelete_key": "http_status_code",
                       "httpdelete_value": "replace_if:200:201"}

        @webob.dec.wsgify
        def app(req):
            return webob.Response(body=json.dumps(app_body), status=200)
        test_filter = edit_response.filter_factory(test_status)(app)

        resp = test_filter(webob.Request.blank("/sauce/id", method="DELETE"))
        self.assertEqual(resp.status_code, 201, resp)



# Copyright 2015 Rackspace Hosting
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import webob.dec
import webob.exc

import wafflehaus.base
import wafflehaus.resource_filter as rf


class BlockResource(wafflehaus.base.WafflehausBase):

    def __init__(self, app, conf):
        super(BlockResource, self).__init__(app, conf)
        self.log.name = conf.get('log_name', __name__)
        self.log.info('Starting wafflehaus resource blocker middleware')
        self.resources = rf.parse_resources(conf.get('resource'))

    def _override(self, req):
        super(BlockResource, self)._override(req)
        new_resource = self._reconf(req, 'str', 'resource')
        if new_resource is not None:
            self.resources = rf.parse_resources(new_resource)

    @webob.dec.wsgify
    def __call__(self, req):
        super(BlockResource, self).__call__(req)
        if not self.enabled:
            return self.app

        if rf.matched_request(req, self.resources):
            self.log.info("Blocked " + str(req.path))
            return webob.exc.HTTPForbidden()
        return self.app


def filter_factory(global_conf, **local_conf):
    """Returns a WSGI filter app for use with paste.deploy."""
    conf = global_conf.copy()
    conf.update(local_conf)

    def block_resource(app):
        return BlockResource(app, conf)
    return block_resource



# Copyright 2015 Rackspace Hosting
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import mock
import webob.exc

from wafflehaus.resource_filter import block_resource
from wafflehaus import tests


class TestResourceFilter(tests.TestCase):
    def setUp(self):
        self.app = mock.Mock()

        self.simple_conf1 = {'resource': 'PoST /widget', 'enabled': 'true'}
        self.simple_conf2 = {'resource': 'PoSt GeT /widget', 'enabled': 'true'}
        self.multi_conf = {'resource': 'post GET /widget, GET posT /derp',
                           'enabled': 'true'}
        self.collapse_conf = {'resource': 'posT /widget, GET /widget',
                              'enabled': 'true'}
        self.complex_conf = {'resource': 'posT /widget/{id}/sub/{sub_id}',
                             'enabled': 'true'}
        self.format_conf1 = {'resource': 'POST /widget{.format:json|xml}',
                             'enabled': 'true'}

    def test_default_instance_create_simple(self):
        result = block_resource.filter_factory(self.simple_conf1)(self.app)
        self.assertIsNotNone(result)
        self.assertTrue(hasattr(result, 'resources'))
        self.assertTrue(isinstance(result.resources, dict))
        self.assertEqual(1, len(result.resources))
        resources = result.resources
        self.assertTrue('/widget' in resources)
        self.assertEqual(1, len(resources['/widget']))

    def test_default_instance_create_simple_multi_method(self):
        result = block_resource.filter_factory(self.simple_conf2)(self.app)
        resources = result.resources
        widget = resources['/widget']
        self.assertEqual(2, len(widget))
        self.assertTrue('POST' in widget)
        self.assertTrue('GET' in widget)

    def test_default_instance_create_multi(self):
        result = block_resource.filter_factory(self.multi_conf)(self.app)
        resources = result.resources
        self.assertEqual(2, len(resources))
        for k, res in resources.iteritems():
            self.assertEqual(2, len(res))
            self.assertTrue('POST' in res)
            self.assertTrue('GET' in res)

    def test_default_instance_collapse(self):
        result = block_resource.filter_factory(self.collapse_conf)(self.app)
        resources = result.resources
        self.assertEqual(1, len(resources))
        widget = resources['/widget']
        self.assertEqual(2, len(widget))
        self.assertTrue('POST' in widget)
        self.assertTrue('GET' in widget)

    def test_match_route(self):
        result = block_resource.filter_factory(self.simple_conf1)(self.app)
        resp = result.__call__.request('/widget', method='POST')
        self.assertTrue(isinstance(resp, webob.exc.HTTPException))

    def test_match_formatted_route1(self):
        result = block_resource.filter_factory(self.format_conf1)(self.app)
        resp = result.__call__.request('/cog', method='POST')
        self.assertEqual(self.app, resp)

        resp = result.__call__.request('/widget', method='POST')
        self.assertTrue(isinstance(resp, webob.exc.HTTPException))
        resp = result.__call__.request('/widget', method='PUT')
        self.assertEqual(self.app, resp)

        resp = result.__call__.request('/widget.json', method='POST')
        self.assertTrue(isinstance(resp, webob.exc.HTTPException))
        resp = result.__call__.request('/widget.json', method='PUT')
        self.assertEqual(self.app, resp)

        resp = result.__call__.request('/widget.xml', method='POST')
        self.assertTrue(isinstance(resp, webob.exc.HTTPException))
        resp = result.__call__.request('/widget.xml', method='PUT')
        self.assertEqual(self.app, resp)

        resp = result.__call__.request('/widget.derp', method='POST')
        self.assertEqual(self.app, resp)
        resp = result.__call__.request('/widget.derp', method='PUT')
        self.assertEqual(self.app, resp)

    def test_match_multi_route(self):
        result = block_resource.filter_factory(self.multi_conf)(self.app)
        resp = result.__call__.request('/widget', method='POST')
        self.assertTrue(isinstance(resp, webob.exc.HTTPException))
        resp = result.__call__.request('/derp', method='POST')
        self.assertTrue(isinstance(resp, webob.exc.HTTPException))
        resp = result.__call__.request('/widget', method='GET')
        self.assertTrue(isinstance(resp, webob.exc.HTTPException))
        resp = result.__call__.request('/derp', method='GET')
        self.assertTrue(isinstance(resp, webob.exc.HTTPException))
        resp = result.__call__.request('/widget', method='PUT')
        self.assertEqual(self.app, resp)
        resp = result.__call__.request('/derp', method='PUT')
        self.assertEqual(self.app, resp)

    def test_match_complex_route(self):
        result = block_resource.filter_factory(self.complex_conf)(self.app)
        resp = result.__call__.request('/widget', method='POST')
        self.assertEqual(self.app, resp)
        resp = result.__call__.request('/widget/1234/sub/1234', method='POST')
        self.assertTrue(isinstance(resp, webob.exc.HTTPException))

    def test_fail_to_match_route(self):
        result = block_resource.filter_factory(self.simple_conf1)(self.app)
        resp = result.__call__.request('/willfail', method='POST')
        self.assertEqual(self.app, resp)

    def test_fail_to_match_method(self):
        result = block_resource.filter_factory(self.simple_conf1)(self.app)
        resp = result.__call__.request('/widget', method='GET')
        self.assertEqual(self.app, resp)

    def test_override_runtime(self):
        self.set_reconfigure()
        result = block_resource.filter_factory(self.simple_conf1)(self.app)
        headers = {'X_WAFFLEHAUS_BLOCKRESOURCE_ENABLED': False}
        resp = result.__call__.request('/widget', method='POST',
                                       headers=headers)
        self.assertEqual(self.app, resp)
        headers = {'X_WAFFLEHAUS_BLOCKRESOURCE_ENABLED': True}
        resp = result.__call__.request('/widget', method='POST',
                                       headers=headers)
        self.assertTrue(isinstance(resp, webob.exc.HTTPException))
        headers = {'X_WAFFLEHAUS_BLOCKRESOURCE_RESOURCE': 'GET /derp'}
        resp = result.__call__.request('/widget', method='POST',
                                       headers=headers)
        self.assertEqual(self.app, resp)
        resp = result.__call__.request('/derp', method='GET',
                                       headers=headers)
        self.assertTrue(isinstance(resp, webob.exc.HTTPException))



# Copyright 2013 Openstack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import json

__all__ = ['set_default_payload', 'get_defaults']


def _walk_json_payload(payload, value, path, depth=0):
    key = path[depth]
    if len(path) - 1 == depth:
        if key not in payload:
            if value == 'null':
                value = None
            payload[key] = value
        return
    if key not in payload:
        return
    ptr = payload[key]
    if isinstance(ptr, list) and not isinstance(ptr, str):
        for child in ptr:
            _walk_json_payload(child, value, path, depth=depth + 1)
    else:
        _walk_json_payload(ptr, value, path, depth=depth + 1)


def json_set_unset_keys(payload, defaults):
    try:
        payload_json = json.loads(payload)
    except ValueError:
        return payload
    for default in defaults:
        _walk_json_payload(payload_json, default['value'], default['path'])
    return json.dumps(payload_json)


def get_defaults(defaults):
    result = []
    if not defaults:
        return result
    work_list = [s.strip() for s in defaults.split(',')]
    for working in work_list:
        parts = [s.strip() for s in working.split('=')]
        path = [s.strip() for s in parts[0].split(':')]
        value = parts[1]
        element = {}
        element['path'] = path
        element['value'] = value
        result.append(element)
    return result



# Copyright 2015 Rackspace Hosting
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import mock
import webob.exc

from wafflehaus.resource_filter import alias
from wafflehaus import tests


class TestResourceFilter(tests.TestCase):
    def setUp(self):
        self.app = mock.Mock()
        self.conf1 = {'resource': '/widget', 'enabled': 'true',
                      'action': ''}
        self.conf2 = {'resource': '/widget', 'enabled': 'true',
                      'action': '301:/cog'}
        self.conf3 = {'resource': '/widget', 'enabled': 'true',
                      'action': 'addslash'}
        self.conf4 = {'resource': '/widget', 'enabled': 'true',
                      'action': 'subrequest:/cog'}

    def test_default_instance_create_simple(self):
        result = alias.filter_factory(self.conf1)(self.app)
        self.assertIsNotNone(result)
        self.assertTrue(hasattr(result, 'resources'))
        self.assertTrue(isinstance(result.resources, dict))
        self.assertEqual(1, len(result.resources))
        resources = result.resources
        self.assertTrue('/widget' in resources)
        widget = resources['/widget']
        self.assertTrue('POST' in widget)
        self.assertTrue('GET' in widget)
        self.assertTrue('PUT' in widget)
        self.assertTrue('DELETE' in widget)
        self.assertTrue('HEAD' in widget)
        self.assertTrue('OPTION' in widget)

    def test_400_on_resource(self):
        result = alias.filter_factory(self.conf1)(self.app)
        resp = result.__call__.request('/widget', method='POST')
        self.assertTrue(isinstance(resp, webob.exc.HTTPBadRequest))
        resp = result.__call__.request('/widget', method='GET')
        self.assertTrue(isinstance(resp, webob.exc.HTTPBadRequest))
        resp = result.__call__.request('/widget', method='PUT')
        self.assertTrue(isinstance(resp, webob.exc.HTTPBadRequest))
        resp = result.__call__.request('/widget', method='DELETE')
        self.assertTrue(isinstance(resp, webob.exc.HTTPBadRequest))
        resp = result.__call__.request('/widget', method='HEAD')
        self.assertTrue(isinstance(resp, webob.exc.HTTPBadRequest))
        resp = result.__call__.request('/widget', method='OPTION')
        self.assertTrue(isinstance(resp, webob.exc.HTTPBadRequest))

        resp = result.__call__.request('/widgets', method='POST')
        self.assertEqual(self.app, resp)
        resp = result.__call__.request('/widgets', method='GET')
        self.assertEqual(self.app, resp)
        resp = result.__call__.request('/widgets', method='PUT')
        self.assertEqual(self.app, resp)
        resp = result.__call__.request('/widgets', method='DELETE')
        self.assertEqual(self.app, resp)
        resp = result.__call__.request('/widgets', method='HEAD')
        self.assertEqual(self.app, resp)
        resp = result.__call__.request('/widgets', method='OPTION')
        self.assertEqual(self.app, resp)

    def test_301_on_resource(self):
        result = alias.filter_factory(self.conf2)(self.app)
        resp = result.__call__.request('/widget', method='POST')
        self.assertTrue(isinstance(resp, webob.exc.HTTPMovedPermanently))
        resp = result.__call__.request('/widget', method='GET')
        self.assertTrue(isinstance(resp, webob.exc.HTTPMovedPermanently))
        resp = result.__call__.request('/widget', method='PUT')
        self.assertTrue(isinstance(resp, webob.exc.HTTPMovedPermanently))
        resp = result.__call__.request('/widget', method='DELETE')
        self.assertTrue(isinstance(resp, webob.exc.HTTPMovedPermanently))
        resp = result.__call__.request('/widget', method='HEAD')
        self.assertTrue(isinstance(resp, webob.exc.HTTPMovedPermanently))
        resp = result.__call__.request('/widget', method='OPTION')
        self.assertTrue(isinstance(resp, webob.exc.HTTPMovedPermanently))

        resp = result.__call__.request('/widgets', method='POST')
        self.assertEqual(self.app, resp)
        resp = result.__call__.request('/widgets', method='GET')
        self.assertEqual(self.app, resp)
        resp = result.__call__.request('/widgets', method='PUT')
        self.assertEqual(self.app, resp)
        resp = result.__call__.request('/widgets', method='DELETE')
        self.assertEqual(self.app, resp)
        resp = result.__call__.request('/widgets', method='HEAD')
        self.assertEqual(self.app, resp)
        resp = result.__call__.request('/widgets', method='OPTION')
        self.assertEqual(self.app, resp)

    def test_add_slash_on_resource(self):
        result = alias.filter_factory(self.conf3)(self.app)
        resp = result.__call__.request('/widget', method='POST')
        self.assertTrue(isinstance(resp, webob.exc.HTTPMovedPermanently))
        resp = result.__call__.request('/widget', method='GET')
        self.assertTrue(isinstance(resp, webob.exc.HTTPMovedPermanently))
        resp = result.__call__.request('/widget', method='PUT')
        self.assertTrue(isinstance(resp, webob.exc.HTTPMovedPermanently))
        resp = result.__call__.request('/widget', method='DELETE')
        self.assertTrue(isinstance(resp, webob.exc.HTTPMovedPermanently))
        resp = result.__call__.request('/widget', method='HEAD')
        self.assertTrue(isinstance(resp, webob.exc.HTTPMovedPermanently))
        resp = result.__call__.request('/widget', method='OPTION')
        self.assertTrue(isinstance(resp, webob.exc.HTTPMovedPermanently))

        resp = result.__call__.request('/widgets', method='POST')
        self.assertEqual(self.app, resp)
        resp = result.__call__.request('/widgets', method='GET')
        self.assertEqual(self.app, resp)
        resp = result.__call__.request('/widgets', method='PUT')
        self.assertEqual(self.app, resp)
        resp = result.__call__.request('/widgets', method='DELETE')
        self.assertEqual(self.app, resp)
        resp = result.__call__.request('/widgets', method='HEAD')
        self.assertEqual(self.app, resp)
        resp = result.__call__.request('/widgets', method='OPTION')
        self.assertEqual(self.app, resp)



# Copyright 2013 Openstack Foundation # All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import logging

import webob
import webob.dec

import wafflehaus.base

log = logging.getLogger(__name__)


class RoleRouter(wafflehaus.base.WafflehausBase):
    """RoleRouter

    The purpose of this class is to route filters based on the role
    obtained from keystone.context.
    """

    def __init__(self, loader, conf):
        self.context_key = conf.get('context_key', 'nova.context')
        self.roles = {}
        self.routes = {}
        self.sorted_routes = []
        self.route_to_role = {}

        # This assumes roles are labeled role_<role> in conf
        routes = conf["routes"].split()
        for route in routes:
            self.sorted_routes.append(route)
            key = "roles_%s" % route
            if key in conf:
                roles = conf[key].split()
                for role in roles:
                    if route not in self.route_to_role:
                        self.route_to_role[route] = []
                    self.route_to_role[route].append(role)
                    self.roles[role] = route

        # This returns the particular route's app after applying the filters
        routes.append("default")
        for route in routes:
            key = "route_%s" % route
            if key in conf:
                pipeline = conf[key]
                pipeline = pipeline.split()
                filters = [loader.get_filter(f) for f in pipeline[:-1]]
                app = loader.get_app(pipeline[-1])
                filters.reverse()
                for f in filters:
                    app = f(app)
                self.routes[route] = app
                if key == 'default':
                    self.app = app
        super(RoleRouter, self).__init__(app, conf)

    @webob.dec.wsgify(RequestClass=webob.Request)
    def __call__(self, req):
        super(RoleRouter, self).__call__(req)
        if not self.enabled:
            return self.app

        context = req.environ.get(self.context_key)

        if not context:
            log.info("No context found")
            return self.routes["default"]

        roles = context.roles
        for current_route in self.sorted_routes:
            route_roles = self.route_to_role[current_route]
            for role in route_roles:
                if role in roles:
                    return self.routes[current_route]

        return self.routes["default"]


def rolerouter_factory(loader, global_conf, **local_conf):
    """Returns a WSGI composite app for use with paste.deploy."""
    conf = global_conf.copy()
    conf.update(local_conf)

    return RoleRouter(loader, conf)



# Copyright 2013 Openstack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from routes import Mapper

__all__ = ['matched_request', 'parse_resources']
_methods = ['POST', 'GET', 'DELETE', 'PUT', 'OPTION', 'HEAD']


def parse_resources(resources):
    result = {}
    if not resources:
        return
    work_list = [s.strip() for s in resources.split(',')]
    for res in work_list:
        res_split = res.split()
        methods = [m.upper() for m in res_split[:-1]]
        resource = res_split[-1]
        if resource not in result:
            result[resource] = []
        if len(methods) == 0:
            result[resource].extend(_methods)
        else:
            for m in methods:
                result[resource].append(m)
    return result


def matched_request(request, resource_list):
    if len(resource_list) == 0:
        return False
    map = Mapper()
    for resource, data in resource_list.iteritems():
        map.connect(None, resource, controller=','.join(data))
    res = map.routematch(request.path)
    if res is None:
        return False
    if request.method in res[0]['controller'].split(','):
        return True
    return False



# Copyright 2013 Openstack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import mock
import webob.exc

from wafflehaus import tests
from wafflehaus.try_context import context_filter


class TestContextClass(context_filter.BaseContextStrategy):
    def __init__(self, key, req_auth=False):
        super(TestContextClass, self).__init__(key, req_auth)
        self._mockme()
        if req_auth:
            self._auth_required()

    def _mockme(self):
        """Exists because one can't mock __init__."""
        pass

    def _auth_required(self):
        """Exists because one can't mock __init__."""
        pass


class TestTryContext(tests.TestCase):

    def setUp(self):
        super(TestTryContext, self).setUp()
        self.app = mock.Mock()
        self.app.return_value = "OK"
        self.start_response = mock.Mock()
        self.test_cls = "wafflehaus.tests.test_try_context.TestContextClass"
        self.context_init = self.create_patch(self.test_cls + '._mockme')
        self.auth_check = self.create_patch(self.test_cls + '._auth_required')

        self.strat_test_auth = {"context_strategy": self.test_cls,
                                'enabled': 'true',
                                'require_auth_info': True,
                                "context_key": "context.test", }

        self.strat_test = {"context_strategy": self.test_cls,
                           'enabled': 'true',
                           "context_key": "context.test", }

        self.strat_testing = {"context_strategy": self.test_cls,
                              "context_key": "context.test",
                              'enabled': 'true',
                              "testing": "true"}

        self.unknown = {"context_strategy": "unknown.class",
                        'enabled': 'true',
                        "context_key": "context.test", }

        self.unknown_testing = {"context_strategy": "unknown.class",
                                "context_key": "context.test",
                                'enabled': 'true',
                                "testing": "true"}
        self.get_admin_mock = mock.Mock()
        self.get_admin_mock.return_value = ['admin']

    def test_create_strategy_test(self):
        """This is a test strategy to see if this thing works."""
        result = context_filter.filter_factory(self.strat_test)(self.app)
        self.assertTrue(isinstance(result, context_filter.ContextFilter))
        resp = result.__call__.request('/something', method='HEAD')
        self.assertEqual(self.app, resp)
        self.assertEqual(1, self.context_init.call_count)
        self.assertEqual(0, self.auth_check.call_count)
        self.assertTrue(isinstance(result, context_filter.ContextFilter))
        """Should be 1 because of a new instance."""
        self.assertEqual(1, self.context_init.call_count)

    def test_create_strategy_test_required_auth(self):
        """This is a test strategy to see if this thing works."""
        result = context_filter.filter_factory(self.strat_test_auth)(self.app)
        self.assertTrue(isinstance(result, context_filter.ContextFilter))
        resp = result.__call__.request('/something', method='HEAD')
        self.assertEqual(self.app, resp)
        self.assertEqual(1, self.context_init.call_count)
        self.assertEqual(1, self.auth_check.call_count)
        self.assertTrue(isinstance(result, context_filter.ContextFilter))
        """Should be 1 because of a new instance."""
        self.assertEqual(1, self.context_init.call_count)

    def test_create_strategy_test_noop_testing(self):
        """This is a test strategy to see if this thing works."""
        result = context_filter.filter_factory(self.strat_test)(self.app)
        self.assertTrue(isinstance(result, context_filter.ContextFilter))
        resp = result.__call__.request('/something', method='HEAD')
        self.assertEqual(self.app, resp)
        self.assertEqual(1, self.context_init.call_count)
        self.assertTrue(isinstance(result, context_filter.ContextFilter))
        """Should be 1 because of a new instance."""
        self.assertEqual(1, self.context_init.call_count)

    def test_create_strategy_none(self):
        """The none strategy simply is a noop."""
        result = context_filter.filter_factory({})(self.app)
        self.assertIsNotNone(result)
        resp = result.__call__.request('/something', method='HEAD')
        self.assertTrue(isinstance(result, context_filter.ContextFilter))
        self.assertEqual(self.app, resp)

    def test_create_strategy_unknown_should_500(self):
        """The none strategy simply is a noop."""
        result = context_filter.filter_factory(self.unknown)(self.app)
        self.assertIsNotNone(result)
        self.assertTrue(isinstance(result, context_filter.ContextFilter))
        resp = result.__call__.request('/something', method='HEAD')
        self.assertNotEqual(self.app, resp)
        self.assertTrue(isinstance(resp, webob.exc.HTTPInternalServerError))

    def test_create_strategy_unknown_should_500_noop_testing(self):
        """The none strategy simply is a noop."""
        result = context_filter.filter_factory(self.unknown_testing)(self.app)
        self.assertIsNotNone(result)
        self.assertTrue(isinstance(result, context_filter.ContextFilter))
        resp = result.__call__.request('/something', method='HEAD')
        self.assertEqual(self.app, resp)

    def test_override_runtime_enabled(self):
        """The none strategy simply is a noop."""
        self.set_reconfigure()
        headers = {'X_WAFFLEHAUS_CONTEXTFILTER_ENABLED': False}
        result = context_filter.filter_factory(self.strat_test)(self.app)
        self.assertTrue(isinstance(result, context_filter.ContextFilter))
        resp = result.__call__.request('/something', method='HEAD',
                                       headers=headers)
        self.assertEqual(self.app, resp)
        self.assertEqual(0, self.context_init.call_count)
        self.assertTrue(isinstance(result, context_filter.ContextFilter))
        """Should be 1 because of a new instance."""
        self.assertEqual(0, self.context_init.call_count)



# Copyright 2013 Openstack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import logging

from oslo_config import cfg
import webob.request

GLOBAL_CONF = cfg.CONF

wafflehaus_global_opts = [
    cfg.BoolOpt('runtime_reconfigurable', default=False,
                help='Will enable header reconfiguration'),
]

GLOBAL_CONF.register_opts(wafflehaus_global_opts, 'WAFFLEHAUS')


class WafflehausBase(object):

    def __init__(self, app, conf):
        self.conf = conf
        self.app = app
        self.header_prefix = 'X_WAFFLEHAUS'
        self.truths = (True, 'True', 'true', 't', '1', 'on', 'yes', 'y')
        logname = __name__
        self.log = logging.getLogger(conf.get('log_name', logname))
        self.testing = (conf.get('testing') in self.truths)
        self.enabled = (conf.get('enabled', False) in self.truths)
        self.reconfigure = GLOBAL_CONF.WAFFLEHAUS.runtime_reconfigurable

    def _reconf(self, req, type, header_suffix, default=None):
        name = self.__class__.__name__.upper()
        ret = default
        header = "%s_%s_%s" % (self.header_prefix, name, header_suffix.upper())
        if header in req.headers:
            val = req.headers[header]
            if type == 'bool':
                ret = val in self.truths
            if type == 'str':
                ret = val
        return ret

    def _override(self, req):
        if not isinstance(req, webob.request.BaseRequest):
            """Ensure that the request is not a mock"""
            return
        self.enabled = self._reconf(req, 'bool', 'enabled', self.enabled)
        self.testing = self._reconf(req, 'bool', 'testing', self.testing)

    def _override_caller(self, req):
        if not self.reconfigure:
            return
        self._override(req)

    def __call__(self, req):
        self._override_caller(req)



# Copyright 2016 Justin Hammond
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import logging
import time
import uuid

import webob.dec
import webob.exc

import wafflehaus.base


FAKE_REQ_ID = 'req-XXXXXXXX-XXXX-XXXX-XXXX-XXXXXXXXXXXX'
FAKE_TENANT_ID = '-'
DEFAULTS = {'request': FAKE_REQ_ID, 'tenant': FAKE_TENANT_ID}


class RequestResponseLogger(wafflehaus.base.WafflehausBase):

    def _get_new_log(self, name, log_file, log_format):
        formatter = logging.Formatter(log_format)
        fileHandler = logging.FileHandler(log_file)
        fileHandler.setFormatter(formatter)

        log = logging.getLogger(name)
        log.handlers = []
        log.addHandler(fileHandler)
        log.propagate = False
        return log

    def __init__(self, app, conf):
        super(RequestResponseLogger, self).__init__(app, conf)
        self.log.name = conf.get('log_name', __name__)
        self.context_key = conf.get('context_key', 'nokey')
        self.separator = ' '
        self.do_detail_logs = (conf.get('do_detail') in self.truths)
        self.detail_level = int(conf.get('detail_level', 500))
        default_format = '%(message)s' + self.separator + '%(asctime)s'
        log_format = default_format
        log_file = conf.get('log_file')

        if log_file is not None:
            self.log = self._get_new_log(self.log.name, log_file, log_format)
            self.log.info('Starting wafflehaus request/response logger')
            if self.do_detail_logs:
                dlog_file = "%s.detail.log" % log_file.replace('.log', '')
                name = "%s.details"
                self.dlog = self._get_new_log(name, dlog_file, '%(message)s')
        else:
            self.enabled = False

    def _log_simple_request(self, req, resp, delta, log_time):
        contents = []
        context_info = DEFAULTS
        ctx = req.environ.get(self.context_key)
        if ctx is not None:
            context_info['request'] = ctx.request_id
            if ctx.tenant_id is not None:
                context_info['tenant'] = ctx.tenant_id
        id = str(uuid.uuid4())
        if context_info['request'] != FAKE_REQ_ID:
            id = context_info['request']

        contents.append("%d" % resp.status_int)
        contents.append("%f sec" % delta)
        contents.append('<--')
        contents.append(req.method)
        qs = ""
        if len(req.query_string) > 0:
            qs = "?%s" % req.query_string
        contents.append("%s%s" % (req.path, qs))
        contents.append(context_info['request'])
        contents.append(context_info['tenant'])

        if self.do_detail_logs:
            contents.append(id)

        """Turn everything into a string."""
        string_contents = [str(c) for c in contents]

        log_str = self.separator.join(string_contents)
        self.log.info(log_str)
        if resp.status_int >= self.detail_level and self.do_detail_logs:
            self._log_detail_request(id, req, resp, delta, log_time)

    def _log_detail_request(self, id, req, resp, delta, log_time):
        context_info = DEFAULTS
        ctx = req.environ.get(self.context_key)
        if ctx is not None:
            context_info['request'] = ctx.request_id
            if ctx.tenant_id is not None:
                context_info['tenant'] = ctx.tenant_id
        qs = ""
        if len(req.query_string) > 0:
            qs = "?%s" % req.query_string
        self.dlog.info("%s | ---" % id)
        self.dlog.info("%s | REQTIME: %s" % (id, log_time))
        self.dlog.info("%s | REQDIFF: %f seconds" % (id, delta))
        self.dlog.info("%s | REQCALL: %s %s%s" % (id, req.method, req.path,
                                                  qs))

        if req.body and len(req.body) > 0:
            body_list = req.body.split('\n')
            for line in body_list:
                self.dlog.info("%s | REQBODY: %s" % (id, line))

        for header, value in req.headers.iteritems():
            self.dlog.info("%s | REQHEAD: %s:%s" % (id, header, value))
        self.dlog.info("%s | REQADDR: %s" % (id, req.remote_addr))

        self.dlog.info("%s | RESTEXT: %s" % (id, resp.status))
        self.dlog.info("%s | RESCODE: %s" % (id, resp.status_int))

        if resp.body and len(resp.body) > 0:
            body_list = resp.body.split('\n')
            for line in body_list:
                self.dlog.info("%s | RESBODY: %s" % (id, line))

    @webob.dec.wsgify
    def __call__(self, req):
        super(RequestResponseLogger, self).__call__(req)
        if not self.enabled:
            return self.app

        start = time.time()
        resp = req.get_response(self.app)
        difference = (time.time() - start)

        log_time = time.strftime("%Y-%m-%d %H:%M")
        try:
            self._log_simple_request(req, resp, difference, log_time)
        except Exception as e:
            self.log("Failed to log due to exception: %s" % e)

        if resp is None:
            return self.app
        else:
            return resp


def filter_factory(global_conf, **local_conf):
    """Returns a WSGI filter app for use with paste.deploy."""
    conf = global_conf.copy()
    conf.update(local_conf)

    def req_resp_logger(app):
        return RequestResponseLogger(app, conf)
    return req_resp_logger



# Copyright 2013 Openstack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import dns.exception
import mock
import webob.exc

from wafflehaus.dns_filter import whitelist
from wafflehaus import tests


def do_lookup(address):
    if address == '192.168.1.1':
        return 'derp.widget.com'
    return 'something.bad.com'


def do_lookup_fail(address):
    if address == '192.168.1.1':
        return 'something.bad.com'
    return 'something.bad.com'


def do_lookup_unknown(address):
    raise dns.exception.DNSException


class FakeARecord(object):

    def __init__(self, good=True):
        if good:
            self.rrset = ['192.168.1.1']
        else:
            self.rrset = ['10.0.0.1']


class FakeResolver(object):

    def __init__(self, name=True):
        self.name = name

    def query(self, value, record_type):
        if record_type == 'PTR':
            return ['derp.widget.com'] if self.name else ['yo.bad.com']
        if record_type == 'A':
            if 'widget.com' in value:
                return FakeARecord()
            else:
                return FakeARecord(good=False)


class TestDNSFilter(tests.TestCase):

    def setUp(self):
        self.app = mock.Mock()
        self.conf_disabled = {'whitelist': 'widget.com'}
        self.conf = {'whitelist': 'widget.com', 'enabled': 'true'}
        self.testconf = {'whitelist': 'widget.com', 'testing': 'true',
                         'enabled': 'true'}
        self.mod_path = 'wafflehaus.dns_filter.whitelist.DNSWhitelist'
        self.addr_path = '%s.get_remote_addr' % self.mod_path
        self.resolver_path = '%s._create_resolver' % self.mod_path
        self.dns_reverse = 'dns.reversename.from_address'

        self.good_ip = '192.168.1.1'
        self.bad_ip = '10.0.0.1'

    def test_create_dns_filter(self):
        result = whitelist.filter_factory(self.conf)(self.app)
        self.assertIsNotNone(result)

    def test_create_dns_filter_not_enabled_by_default(self):
        result = whitelist.filter_factory(self.conf_disabled)(self.app)
        self.assertIsNotNone(result)
        self.assertFalse(result.enabled)

    def test_match_ok(self):
        result = whitelist.filter_factory(self.conf)(self.app)
        m_addr = self.create_patch(self.addr_path)
        m_addr.return_value = self.good_ip

        m_resolve = self.create_patch(self.resolver_path)
        m_resolve.return_value = FakeResolver()

        m_dns_rname = self.create_patch(self.dns_reverse)
        m_dns_rname.side_effect = ['omg.widget.com']

        resp = result.__call__.request('/widget', method='POST')
        self.assertEqual(1, m_addr.call_count)
        self.assertEqual(1, m_dns_rname.call_count)
        self.assertEqual(1, m_resolve.call_count)
        self.assertFalse(isinstance(resp, webob.exc.HTTPForbidden))

    def test_match_ok_with_forwarded_header(self):
        result = whitelist.filter_factory(self.conf)(self.app)
        m_addr = self.create_patch(self.addr_path)
        m_addr.return_value = self.bad_ip

        m_resolve = self.create_patch(self.resolver_path)
        m_resolve.return_value = FakeResolver()

        m_dns_rname = self.create_patch(self.dns_reverse)
        m_dns_rname.side_effect = ['omg.widget.com']

        ip_list = [self.good_ip, self.bad_ip]
        headers = {'X-Forwarded-For': ','.join(ip_list)}

        resp = result.__call__.request('/widget', method='POST',
                                       headers=headers)
        self.assertEqual(1, m_addr.call_count)
        self.assertEqual(1, m_dns_rname.call_count)
        self.assertEqual(1, m_resolve.call_count)
        self.assertFalse(isinstance(resp, webob.exc.HTTPForbidden))

# TESTING FUNCTIONS

    def test_no_fail_on_name_lookup_while_testing(self):
        result = whitelist.filter_factory(self.testconf)(self.app)
        m_addr = self.create_patch(self.addr_path)
        m_addr.return_value = self.good_ip

        m_resolve = self.create_patch(self.resolver_path)
        m_resolve.return_value = FakeResolver(name=False)

        m_dns_rname = self.create_patch(self.dns_reverse)
        m_dns_rname.side_effect = ['omg.widget.com']

        resp = result.__call__.request('/widget', method='POST')
        self.assertEqual(1, m_addr.call_count)
        self.assertEqual(1, m_dns_rname.call_count)
        self.assertEqual(1, m_resolve.call_count)
        self.assertFalse(isinstance(resp, webob.exc.HTTPForbidden))

    def test_no_fail_on_match_bad_address_while_testing(self):
        result = whitelist.filter_factory(self.testconf)(self.app)
        m_addr = self.create_patch(self.addr_path)
        m_addr.return_value = self.bad_ip

        m_resolve = self.create_patch(self.resolver_path)
        m_resolve.return_value = FakeResolver()

        m_dns_rname = self.create_patch(self.dns_reverse)
        m_dns_rname.side_effect = ['omg.widget.com']

        resp = result.__call__.request('/widget', method='POST')
        self.assertEqual(1, m_addr.call_count)
        self.assertEqual(1, m_dns_rname.call_count)
        self.assertEqual(1, m_resolve.call_count)
        self.assertFalse(isinstance(resp, webob.exc.HTTPForbidden))

    def test_no_fail_match_bad_name_while_testing(self):
        result = whitelist.filter_factory(self.testconf)(self.app)
        m_addr = self.create_patch(self.addr_path)
        m_addr.return_value = self.good_ip

        m_resolve = self.create_patch(self.resolver_path)
        m_resolve.return_value = FakeResolver()

        m_dns_rname = self.create_patch(self.dns_reverse)
        m_dns_rname.side_effect = ['something.bad.com']

        resp = result.__call__.request('/widget', method='POST')
        self.assertEqual(1, m_addr.call_count)
        self.assertEqual(1, m_dns_rname.call_count)
        self.assertEqual(1, m_resolve.call_count)
        self.assertFalse(isinstance(resp, webob.exc.HTTPForbidden))

    def test_no_fail_match_unknown_address_while_testing(self):
        result = whitelist.filter_factory(self.testconf)(self.app)
        m_addr = self.create_patch(self.addr_path)
        m_addr.return_value = self.good_ip

        m_resolve = self.create_patch(self.resolver_path)
        m_resolve.return_value = FakeResolver()

        m_dns_rname = self.create_patch(self.dns_reverse)
        m_dns_rname.side_effect = [dns.exception.DNSException]

        resp = result.__call__.request('/widget', method='POST')
        self.assertEqual(1, m_addr.call_count)
        self.assertEqual(1, m_dns_rname.call_count)
        self.assertEqual(1, m_resolve.call_count)
        self.assertFalse(isinstance(resp, webob.exc.HTTPForbidden))

    def test_no_fail_no_dns_entries_while_testing(self):
        result = whitelist.filter_factory(self.testconf)(self.app)
        m_addr = self.create_patch(self.addr_path)
        resp = result.__call__.request('/widget', method='POST')
        self.assertEqual(1, m_addr.call_count)
        self.assertFalse(isinstance(resp, webob.exc.HTTPForbidden))

# END TESTING

    def test_fail_on_name_lookup(self):
        result = whitelist.filter_factory(self.conf)(self.app)
        m_addr = self.create_patch(self.addr_path)
        m_addr.return_value = self.good_ip

        m_resolve = self.create_patch(self.resolver_path)
        m_resolve.return_value = FakeResolver(name=False)

        m_dns_rname = self.create_patch(self.dns_reverse)
        m_dns_rname.side_effect = ['omg.widget.com']

        resp = result.__call__.request('/widget', method='POST')
        self.assertEqual(1, m_addr.call_count)
        self.assertEqual(1, m_dns_rname.call_count)
        self.assertEqual(1, m_resolve.call_count)
        self.assertTrue(isinstance(resp, webob.exc.HTTPForbidden))

    def test_match_bad_address(self):
        result = whitelist.filter_factory(self.conf)(self.app)
        m_addr = self.create_patch(self.addr_path)
        m_addr.return_value = self.bad_ip

        m_resolve = self.create_patch(self.resolver_path)
        m_resolve.return_value = FakeResolver()

        m_dns_rname = self.create_patch(self.dns_reverse)
        m_dns_rname.side_effect = ['omg.widget.com']

        resp = result.__call__.request('/widget', method='POST')
        self.assertEqual(1, m_addr.call_count)
        self.assertEqual(1, m_dns_rname.call_count)
        self.assertEqual(1, m_resolve.call_count)
        self.assertTrue(isinstance(resp, webob.exc.HTTPForbidden))

    def test_match_bad_name(self):
        result = whitelist.filter_factory(self.conf)(self.app)
        m_addr = self.create_patch(self.addr_path)
        m_addr.return_value = self.good_ip

        m_resolve = self.create_patch(self.resolver_path)
        m_resolve.return_value = FakeResolver()

        m_dns_rname = self.create_patch(self.dns_reverse)
        m_dns_rname.side_effect = ['something.bad.com']

        resp = result.__call__.request('/widget', method='POST')
        self.assertEqual(1, m_addr.call_count)
        self.assertEqual(1, m_dns_rname.call_count)
        self.assertEqual(1, m_resolve.call_count)
        self.assertTrue(isinstance(resp, webob.exc.HTTPForbidden))

    def test_match_unknown_address(self):
        result = whitelist.filter_factory(self.conf)(self.app)
        m_addr = self.create_patch(self.addr_path)
        m_addr.return_value = self.good_ip

        m_resolve = self.create_patch(self.resolver_path)
        m_resolve.return_value = FakeResolver()

        m_dns_rname = self.create_patch(self.dns_reverse)
        m_dns_rname.side_effect = [dns.exception.DNSException]

        resp = result.__call__.request('/widget', method='POST')
        self.assertEqual(1, m_addr.call_count)
        self.assertEqual(1, m_dns_rname.call_count)
        self.assertEqual(1, m_resolve.call_count)
        self.assertTrue(isinstance(resp, webob.exc.HTTPForbidden))

    def test_no_dns_entries(self):
        result = whitelist.filter_factory(self.conf)(self.app)
        m_addr = self.create_patch(self.addr_path)
        resp = result.__call__.request('/widget', method='POST')
        self.assertTrue(m_addr.called_once)
        self.assertTrue(isinstance(resp, webob.exc.HTTPForbidden))

    def test_no_whitelist_error(self):
        result = whitelist.filter_factory({'enabled': 'true'})(self.app)
        resp = result.__call__.request('/widget', method='POST')
        self.assertTrue(isinstance(resp, webob.exc.HTTPInternalServerError))

    def test_fail_with_empty_forwarded_header(self):
        result = whitelist.filter_factory(self.conf)(self.app)
        m_addr = self.create_patch(self.addr_path)
        m_addr.return_value = self.bad_ip

        m_resolve = self.create_patch(self.resolver_path)
        m_resolve.return_value = FakeResolver()

        m_dns_rname = self.create_patch(self.dns_reverse)
        m_dns_rname.side_effect = ['omg.widget.com']

        headers = {'X-Forwarded-For': ''}

        resp = result.__call__.request('/widget', method='POST',
                                       headers=headers)
        self.assertEqual(1, m_addr.call_count)
        self.assertEqual(1, m_dns_rname.call_count)
        self.assertEqual(1, m_resolve.call_count)
        self.assertTrue(isinstance(resp, webob.exc.HTTPForbidden))

    def test_runtime_overrides(self):
        self.set_reconfigure()
        result = whitelist.filter_factory(self.conf)(self.app)
        m_addr = self.create_patch(self.addr_path)
        m_addr.return_value = self.good_ip

        m_resolve = self.create_patch(self.resolver_path)
        m_resolve.return_value = FakeResolver()

        m_dns_rname = self.create_patch(self.dns_reverse)
        m_dns_rname.side_effect = ['omg.widget.com']

        headers = {'X_WAFFLEHAUS_DNSWHITELIST_ENABLED': False}

        result.__call__.request('/widget', method='POST',
                                headers=headers)
        self.assertEqual(0, m_addr.call_count)
        self.assertEqual(0, m_dns_rname.call_count)
        self.assertEqual(0, m_resolve.call_count)



# Copyright 2013 Openstack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import json

import mock

from wafflehaus.payload_filter import unset_key
from wafflehaus import tests


class TestPayloadFilter(tests.TestCase):
    def setUp(self):
        self.app = mock.Mock()
        simple_widget = 'widget:thing=thingie'
        complex_widget = 'widget:sub:thing=thingie'
        multi_widget = 'widgets:thing=thingie'
        multi_subwidget = 'widgets:sub:thing=thingie'
        self.simple_conf1 = {'resource': 'POST /widget', 'enabled': 'true',
                             'defaults': simple_widget}
        self.simple_conf2 = {'resource': 'POST /widget', 'enabled': 'true',
                             'defaults': complex_widget}
        self.multi_conf = {'resource': 'POST /widget', 'enabled': 'true',
                           'defaults': '%s,%s' % (simple_widget,
                                                  complex_widget)}
        self.multi_confr = {'resource': 'POST /widget', 'enabled': 'true',
                            'defaults': '%s,%s' % (complex_widget,
                                                   simple_widget)}
        self.plural_conf = {'resource': 'POST /widget', 'enabled': 'true',
                            'defaults': multi_widget}
        self.plural_conf2 = {'resource': 'POST /widget', 'enabled': 'true',
                             'defaults': multi_subwidget}

        self.body1 = '{"widget": { "name": "foo"}}'
        self.body2 = '{"widget": { "name": "foo", "thing": "derp"}}'
        self.body3 = '{"widget": { "name": "foo", "sub": { "name": "bar"}}}'
        self.body4 = '{"widgets": [{"name": "1"},{"name": "2"}]}'
        self.body5 = '{"widgets": [{"sub":{"name": "1"}}]}'

    def test_default_instance_create_simple(self):
        result = unset_key.filter_factory(self.simple_conf1)(self.app)
        self.assertIsNotNone(result)
        self.assertTrue(hasattr(result, 'resources'))
        self.assertTrue(isinstance(result.resources, dict))
        self.assertEqual(1, len(result.resources))
        resources = result.resources
        self.assertTrue('/widget' in resources)
        self.assertEqual(1, len(resources['/widget']))

    def test_request_body_overridden(self):
        """Payload filter will set values for keys that do not exist."""
        result = unset_key.filter_factory(self.simple_conf1)(self.app)
        resp = result.__call__.request('/widget', method='POST',
                                       body=self.body1)
        self.assertEqual(self.app, resp)
        body = result.body
        self.assertIsNotNone(body)
        json_body = json.loads(body)
        self.assertTrue('widget' in json_body)
        self.assertTrue('thing' not in json_body)
        widget = json_body['widget']
        self.assertTrue('name' in widget)
        self.assertEqual('foo', widget['name'])
        self.assertTrue('thing' in widget)
        self.assertEqual('thingie', widget['thing'])

    def test_request_body_not_overridden(self):
        """Payload filter will not change values that are set."""
        result = unset_key.filter_factory(self.simple_conf1)(self.app)
        resp = result.__call__.request('/widget', method='POST',
                                       body=self.body2)
        self.assertEqual(self.app, resp)
        body = result.body
        self.assertIsNotNone(body)
        json_body = json.loads(body)
        self.assertTrue('widget' in json_body)
        self.assertTrue('thing' not in json_body)
        widget = json_body['widget']
        self.assertTrue('name' in widget)
        self.assertEqual('foo', widget['name'])
        self.assertTrue('thing' in widget)
        self.assertEqual('derp', widget['thing'])

    def test_request_complex_path(self):
        result = unset_key.filter_factory(self.simple_conf2)(self.app)
        resp = result.__call__.request('/widget', method='POST',
                                       body=self.body3)
        self.assertEqual(self.app, resp)
        body = result.body
        self.assertIsNotNone(body)
        json_body = json.loads(body)
        self.assertTrue('widget' in json_body)
        self.assertTrue('thing' not in json_body)
        widget = json_body['widget']
        self.assertTrue('thing' not in widget)
        self.assertTrue('name' in widget)
        self.assertEqual('foo', widget['name'])
        self.assertTrue('sub' in widget)
        sub = widget['sub']
        self.assertEqual('thingie', sub['thing'])

    def test_request_multi_path(self):
        result = unset_key.filter_factory(self.multi_conf)(self.app)
        resp = result.__call__.request('/widget', method='POST',
                                       body=self.body3)
        self.assertEqual(self.app, resp)
        body = result.body
        self.assertIsNotNone(body)
        json_body = json.loads(body)
        self.assertTrue('widget' in json_body)
        widget = json_body['widget']
        self.assertTrue('name' in widget)
        self.assertEqual('foo', widget['name'])
        self.assertTrue('thing' in widget)
        self.assertEqual('thingie', widget['thing'])
        self.assertTrue('sub' in widget)
        sub = widget['sub']
        self.assertEqual('thingie', sub['thing'])

    def test_request_multi_path_with_part_missing(self):
        result = unset_key.filter_factory(self.multi_conf)(self.app)
        resp = result.__call__.request('/widget', method='POST',
                                       body=self.body1)
        self.assertEqual(self.app, resp)
        body = result.body
        self.assertIsNotNone(body)
        json_body = json.loads(body)
        self.assertTrue('widget' in json_body)
        widget = json_body['widget']
        self.assertTrue('name' in widget)
        self.assertEqual('foo', widget['name'])
        self.assertTrue('thing' in widget)
        self.assertEqual('thingie', widget['thing'])
        self.assertFalse('sub' in widget)

    def test_request_multi_path_with_part_missing_reversed(self):
        result = unset_key.filter_factory(self.multi_confr)(self.app)
        resp = result.__call__.request('/widget', method='POST',
                                       body=self.body1)
        self.assertEqual(self.app, resp)
        body = result.body
        self.assertIsNotNone(body)
        json_body = json.loads(body)
        self.assertTrue('widget' in json_body)
        widget = json_body['widget']
        self.assertTrue('name' in widget)
        self.assertEqual('foo', widget['name'])
        self.assertTrue('thing' in widget)
        self.assertEqual('thingie', widget['thing'])
        self.assertFalse('sub' in widget)

    def test_request_plural_request(self):
        result = unset_key.filter_factory(self.plural_conf)(self.app)
        resp = result.__call__.request('/widget', method='POST',
                                       body=self.body4)
        self.assertEqual(self.app, resp)
        body = result.body
        self.assertIsNotNone(body)
        json_body = json.loads(body)
        self.assertTrue('widgets' in json_body)
        self.assertTrue('thing' not in json_body)
        widgets = json_body['widgets']
        for widget in widgets:
            self.assertTrue('name' in widget)
            self.assertTrue('thing' in widget)
            self.assertEqual('thingie', widget['thing'])

    def test_request_plural_sub_request(self):
        result = unset_key.filter_factory(self.plural_conf2)(self.app)
        resp = result.__call__.request('/widget', method='POST',
                                       body=self.body5)
        self.assertEqual(self.app, resp)
        body = result.body
        self.assertIsNotNone(body)
        json_body = json.loads(body)
        self.assertTrue('widgets' in json_body)
        self.assertTrue('thing' not in json_body)
        widgets = json_body['widgets']
        for widget in widgets:
            self.assertTrue('thing' not in widget)
            self.assertTrue('sub' in widget)
            sub = widget['sub']
            self.assertTrue('name' in sub)
            self.assertTrue('thing' in sub)
            self.assertEqual('thingie', sub['thing'])

    def test_override_runtime(self):
        self.set_reconfigure()
        result = unset_key.filter_factory(self.plural_conf2)(self.app)
        headers = {'X_WAFFLEHAUS_DEFAULTPAYLOAD_ENABLED': False}
        resp = result.__call__.request('/widget', method='POST',
                                       body=self.body5, headers=headers)
        self.assertEqual(self.app, resp)
        self.assertFalse(hasattr(result, 'body'))
        headers = {'X_WAFFLEHAUS_DEFAULTPAYLOAD_ENABLED': True}
        resp = result.__call__.request('/widget', method='POST',
                                       body=self.body5, headers=headers)
        self.assertEqual(self.app, resp)
        self.assertTrue(hasattr(result, 'body'))



# Copyright 2013 Openstack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""This middleware is intended to be used with paste.deploy."""

import dns.exception
import dns.resolver
import dns.reversename
import webob.dec
import webob.exc

import wafflehaus.base


# pylint: disable=R0903
# pylint: disable=H405
class DNSWhitelist(wafflehaus.base.WafflehausBase):
    """DNSWhitelist middleware.

    Will DNS lookup REMOTE_ADDR and attempt to match result to a whitelist. A
    failed match will 403.
    """
    def __init__(self, app, conf):
        super(DNSWhitelist, self).__init__(app, conf)
        self.log.name = conf.get('log_name', __name__)
        self.log.info('Starting wafflehaus dns whitelist middleware')
        self.ignore_forwarded = (conf.get('ignore_forwarded') in self.truths)
        self.whitelist = self._create_whitelist(conf.get('whitelist'))

    def _create_resolver(self):
        """Creates the DNS resolver."""
        nameserver = str(dns.resolver.Resolver().nameservers[0])
        nameserver = self.conf.get('nameserver', nameserver)
        res = dns.resolver.Resolver(configure=False)
        res.nameservers = [nameserver]
        return res

    def _create_whitelist(self, whitelist):
        """Creates the whitelist from configuration or testing whitelists."""
        result = None
        if self.testing:
            test_list = ['rackspace.com']
            result = test_list
        if whitelist is not None:
            result = whitelist.split(" ")
        return result

    def check_reverse_dns(self, ip_address, a_record_rrset):
        """Checks to ensure IP is within a set of IPs from an A query set."""
        match = any(ip_address == str(val) for val in a_record_rrset)
        return match

    def check_domain_to_whitelist(self, domain):
        self.log.info("Checking " + str(domain))
        if domain.endswith('.'):
            domain = domain[:-1]
        for ok_host in self.whitelist:
            if domain.endswith(ok_host):
                return True
        return False

    def get_remote_addr(self, request):
        return request.remote_addr

    def _override(self, req):
        super(DNSWhitelist, self)._override(req)
        new_whitelist = self._reconf(req, 'str', 'whitelist', None)
        if new_whitelist is not None:
            self.whitelist = self._create_whitelist(new_whitelist)
        self.ignore_forwarded = self._reconf(req, 'bool', 'ignore_forwarded',
                                             self.ignore_forwarded)

    def parse_x_forwarded_for(self, xforward):
        """It is a CSV list of IPs."""
        self.log.info("Forwarded is : " + str(xforward))
        ip_list = xforward.split(',')
        if len(ip_list):
            return ip_list[0]
        return None

    @webob.dec.wsgify
    def __call__(self, req):
        super(DNSWhitelist, self).__call__(req)
        if not self.enabled:
            return self.app

        if not self.whitelist:
            self.log.error("Whitelist not set")
            return webob.exc.HTTPInternalServerError()

        remote_addr = self.get_remote_addr(req)

        if 'X-Forwarded-For' in req.headers and not self.ignore_forwarded:
            remote_addr = self.parse_x_forwarded_for(
                req.headers['X-Forwarded-For'])
            if remote_addr is None:
                return webob.exc.HTTPForbidden()

        if self.testing:
            remote_addr = self.conf.get('testing_remote_addr', remote_addr)

        res = self._create_resolver()

        try:
            name = dns.reversename.from_address(remote_addr)
            ptr = res.query(name, "PTR")[0]

            if not self.check_domain_to_whitelist(str(name)):
                self.log.warning("DNS whitelist matching failure")
                if not self.testing:
                    return webob.exc.HTTPForbidden()
                else:
                    return self.app

            a_record = res.query(str(ptr), "A")
        except dns.exception.DNSException:
            msg = "Missing DNS entries?"
            self.log.error("DNS Error during query: " + msg)
            if not self.testing:
                return webob.exc.HTTPForbidden()
            else:
                return self.app

        if not self.check_reverse_dns(remote_addr, a_record.rrset):
            if self.testing:
                self.log.warning("Reverse DNS check failed")
            else:
                self.log.warning("Reverse DNS check failed")
                return webob.exc.HTTPForbidden()
        return self.app


def filter_factory(global_conf, **local_conf):
    """Returns a WSGI filter app for use with paste.deploy."""
    conf = global_conf.copy()
    conf.update(local_conf)

    def auth_bypass(app):
        """Returns the app for paste.deploy."""
        return DNSWhitelist(app, conf)
    return auth_bypass
